# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class SessionStateData(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'startup': 'SessionStartingUpData',
        'stopping': 'SessionStoppingData'
    }

    attribute_map = {
        'startup': 'startup',
        'stopping': 'stopping'
    }

    def __init__(self, startup=None, stopping=None, local_vars_configuration=None):  # noqa: E501
        """SessionStateData - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._startup = None
        self._stopping = None
        self.discriminator = None

        if startup is not None:
            self.startup = startup
        if stopping is not None:
            self.stopping = stopping

    @property
    def startup(self):
        """Gets the startup of this SessionStateData.  # noqa: E501


        :return: The startup of this SessionStateData.  # noqa: E501
        :rtype: SessionStartingUpData
        """
        return self._startup

    @startup.setter
    def startup(self, startup):
        """Sets the startup of this SessionStateData.


        :param startup: The startup of this SessionStateData.  # noqa: E501
        :type: SessionStartingUpData
        """

        self._startup = startup

    @property
    def stopping(self):
        """Gets the stopping of this SessionStateData.  # noqa: E501


        :return: The stopping of this SessionStateData.  # noqa: E501
        :rtype: SessionStoppingData
        """
        return self._stopping

    @stopping.setter
    def stopping(self, stopping):
        """Sets the stopping of this SessionStateData.


        :param stopping: The stopping of this SessionStateData.  # noqa: E501
        :type: SessionStoppingData
        """

        self._stopping = stopping

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, SessionStateData):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, SessionStateData):
            return True

        return self.to_dict() != other.to_dict()



import dataclasses
import ipaddress
from typing import Union

from anyscale.cli_logger import BlockLogger


@dataclasses.dataclass
class CapacityThreshold:
    min_network_size: Union[ipaddress.IPv4Network, ipaddress.IPv6Network]
    warn_network_size: Union[ipaddress.IPv4Network, ipaddress.IPv6Network]
    resource_type: str

    def verify_network_capacity(
        self, *, cidr_block_str: str, resource_name: str, logger: BlockLogger
    ) -> bool:
        cidr_block = ipaddress.ip_network(cidr_block_str, strict=False)

        min_hosts = self.min_network_size.num_addresses
        warn_hosts = self.warn_network_size.num_addresses

        if cidr_block.num_addresses < min_hosts:
            logger.error(
                f"The provided {self.resource_type} ({resource_name})'s CIDR block ({cidr_block}) is too"
                f" small. We want at least {min_hosts} addresses,"
                f" but this {self.resource_type} only has {cidr_block.num_addresses}. Please reach out to"
                f" support if this is an issue!"
            )
            return False
        elif cidr_block.num_addresses < warn_hosts:
            logger.warning(
                f"The provided {self.resource_type} ({resource_name})'s CIDR block ({cidr_block}) is probably"
                f" too small. We suggest at least {warn_hosts}"
                f" addresses, but this {self.resource_type} only supports up to"
                f" {cidr_block.num_addresses} addresses."
            )

        return True


AWS_VPC_CAPACITY = CapacityThreshold(
    min_network_size=ipaddress.ip_network("10.0.0.0/24"),
    warn_network_size=ipaddress.ip_network("10.0.0.0/20"),
    resource_type="VPC",
)

AWS_SUBNET_CAPACITY = CapacityThreshold(
    min_network_size=ipaddress.ip_network("10.0.0.0/28"),
    warn_network_size=ipaddress.ip_network("10.0.0.0/24"),
    resource_type="Subnet",
)

GCP_SUBNET_CAPACITY = dataclasses.replace(AWS_VPC_CAPACITY, resource_type="Subnet")



import copy
import json
from pathlib import Path
import re
import sys
from threading import Thread
from typing import Any, Callable, Dict, Iterator, List, Optional, Tuple
from unittest.mock import Mock, patch
from wsgiref.simple_server import make_server

from google.api_core.exceptions import NotFound
from google.auth.credentials import AnonymousCredentials
from google.cloud.compute_v1.types import Subnetwork
from googleapiclient.errors import HttpError
import pytest

from anyscale.cli_logger import BlockLogger
from anyscale.gcp_verification import (
    _gcp_subnet_has_enough_capacity,
    _verify_gcp_access_service_account,
    _verify_gcp_dataplane_service_account,
    _verify_gcp_networking,
    _verify_gcp_project,
    GCPLogger,
    GoogleCloudClientFactory,
    verify_gcp,
)


class RequestTracker:
    def __init__(self):
        self.responses: Dict[str, List[Tuple[int, Optional[str]]]] = {}
        self.seen_requests: List[Any] = []

    def reset(self, responses):
        for k, v in responses.items():
            re.compile(k)
            assert all(len(i) == 2 for i in v)
        self.responses = copy.deepcopy(responses)
        self.seen_requests = []


class GCloudMockHandler:
    def __init__(self, tracker, *args, **kwargs):
        self.tracker = tracker

    def __call__(self, env, start_response):
        self.tracker.seen_requests.append(env)

        path = env["PATH_INFO"]

        for regex in self.tracker.responses:
            if re.match(regex, path):
                code, body_file = self.tracker.responses[regex].pop(0)
                start_response(code, [("Content-Type", "application/json")])
                if body_file is not None:
                    with open(
                        Path(__file__).parent.joinpath("gcp_responses", body_file)
                    ) as f:
                        return [json.dumps(json.load(f)).strip().encode()]
                return []

        pytest.fail(
            f"Un handled request to {self.path}:\n{self.tracker.responses}",
            pytrace=False,
        )


@pytest.fixture()
def setup_mock_server() -> Iterator[Tuple[GoogleCloudClientFactory, RequestTracker]]:
    tracker = RequestTracker()
    server = make_server("localhost", 0, GCloudMockHandler(tracker))
    port = server.server_port
    print(f"Serving on (http://localhost:{port})")
    t = Thread(target=server.serve_forever, daemon=True)
    t.start()
    yield (
        GoogleCloudClientFactory(
            credentials=AnonymousCredentials(),
            force_rest=True,
            client_options={"api_endpoint": f"http://127.0.0.1:{port}"},
        ),
        tracker,
    )
    print(f"Shutting down server on (http://localhost:{port})")
    server.server_close()
    t.join(timeout=0.1)


_MOCK_RESOURCES = {
    "vpc_name": "vpc_one",
    "project_id": "anyscale-bridge-deadbeef",
    "anyscale_access_service_account": "anyscale-admin@anyscale-bridge-deadbeef.iam.gserviceaccount.com",
    "dataplane_service_account": "cld-xyz-access@anyscale-bridge-deadbeef.iam.gserviceaccount.com",
    "subnet_name": "subnet_one",
    "region": "us-west1",
}


@pytest.mark.parametrize(
    ("responses", "expected_result", "num_calls"),
    [
        pytest.param(
            {".*/networks/.*": [("404 Not Found", None)]}, False, 1, id="Network404",
        ),
        pytest.param(
            {
                ".*/networks/.*": [("200 OK", "networks.json")],
                ".*/subnetworks/.*": [("404 Not Found", None)],
            },
            False,
            2,
            id="SubNet404",
        ),
        pytest.param(
            {
                ".*/networks/.*": [("200 OK", "networks.json")],
                ".*/subnetworks/.*": [("200 OK", "subnetworks.json")],
            },
            True,
            2,
            id="Success",
        ),
        pytest.param(
            {
                ".*/networks/.*": [("200 OK", "networks.json")],
                ".*/subnetworks/.*": [("200 OK", "subnetworks_other.json")],
            },
            False,
            2,
            id="WrongNetwrork",
        ),
    ],
)
def test_verify_gcp_networking(
    setup_mock_server, responses, expected_result: bool, num_calls: int
):
    factory, tracker = setup_mock_server
    tracker.reset(responses=responses)
    assert (
        _verify_gcp_networking(factory, _MOCK_RESOURCES, _gcp_logger())
        == expected_result
    )
    assert len(tracker.seen_requests) == num_calls, tracker.seen_requests


@pytest.mark.parametrize(
    ("cidr", "result", "log"),
    [
        ("10.0.0.0/16", True, ""),
        ("10.0.0.0/19", True, ""),
        ("10.0.0.0/21", True, "We suggest at least"),
        ("10.0.0.0/25", False, "We want at least"),
        pytest.param(
            "",
            False,
            "",
            marks=pytest.mark.xfail(raises=ValueError, strict=True),
            id="BadIPAddress",
        ),
    ],
)
def test_gcp_subnet_has_enough_capacity(cidr: str, result: bool, log: str, capsys):
    subnet = Subnetwork(ip_cidr_range=cidr)

    assert _gcp_subnet_has_enough_capacity(subnet, BlockLogger()) == result

    stdout, stderr = capsys.readouterr()
    sys.stdout.write(stdout)
    sys.stderr.write(stderr)

    if log:
        assert re.search(log, stderr)


@pytest.mark.parametrize(
    "call_factory",
    [
        pytest.param(
            lambda f: f.compute_v1.NetworksClient().get(project="abc", network="cde"),
            id="Compute",
        ),
        pytest.param(
            lambda f: f.resourcemanager_v3.ProjectsClient().get_project(
                name="projects/abc"
            ),
            id="ResourceManager",
        ),
    ],
)
def test_client_factory_cloud_client(setup_mock_server, call_factory: Callable):
    factory, tracker = setup_mock_server
    tracker.reset(responses={".*": [("404 Not Found", None)]})
    with pytest.raises(NotFound):
        call_factory(factory)


def test_client_factory_apis(setup_mock_server):
    factory, tracker = setup_mock_server
    tracker.reset(responses={".*": [("418 I'm a teapot", None)]})

    with pytest.raises(HttpError) as e:
        factory.build("iam", "v1").projects().serviceAccounts().get(
            name="projects/-/serviceAccounts/abc"
        ).execute()
    assert e.value.status_code == 418


@pytest.mark.parametrize("projects_match", [True, False])
def test_gcp_verify_credentials_project(capsys, projects_match: bool):

    credentials_project = (
        _MOCK_RESOURCES["project_id"] if projects_match else "other_gcp_project"
    )
    with patch(
        "anyscale.gcp_verification.get_application_default_credentials"
    ) as mock_credentials, patch.multiple(
        "anyscale.gcp_verification",
        _verify_gcp_project=Mock(),
        _verify_gcp_access_service_account=Mock(),
        _verify_gcp_dataplane_service_account=Mock(),
        _verify_gcp_networking=Mock(),
    ):
        mock_credentials.return_value = (
            AnonymousCredentials(),
            credentials_project,
        )
        assert verify_gcp(_MOCK_RESOURCES, BlockLogger())
        mock_credentials.assert_called_once()

    _, err = capsys.readouterr()
    assert ("Default credentials are for" in err) != projects_match


@pytest.mark.parametrize(
    ("responses", "expected_result"),
    [
        pytest.param({".*": [("404 Not Found", None)]}, False, id="NotFound"),
        pytest.param(
            {
                ".*": [
                    ("200 OK", "project_get.json"),
                    ("200 OK", "project_iam_bindings.json"),
                ]
            },
            True,
            id="ProjectExists",
        ),
        pytest.param(
            {".*": [("200 OK", "project_get_inactive.json")]},
            False,
            id="ProjectInactive",
        ),
    ],
)
def test_verify_gcp_project(setup_mock_server, responses, expected_result: bool):
    factory, tracker = setup_mock_server
    tracker.reset(responses)
    assert (
        _verify_gcp_project(factory, _MOCK_RESOURCES, _gcp_logger()) == expected_result
    )


def test_verify_gcp_missing_bindigs(setup_mock_server, capsys):
    factory, tracker = setup_mock_server
    tracker.reset(
        {
            ".*": [
                ("200 OK", "project_get.json"),
                ("200 OK", "project_iam_bindings.json"),
            ]
        }
    )
    assert _verify_gcp_project(factory, _MOCK_RESOURCES, _gcp_logger())
    _, err = capsys.readouterr()
    assert (
        "The Compute Engine Service Agent does not have the standard IAM Role of 'roles/compute.serviceAgent'"
        in err
    )


@pytest.mark.parametrize(
    ("service_account", "expected_result"),
    [
        pytest.param(
            "anyscale-admin-owner@anyscale-bridge-deadbeef.iam.gserviceaccount.com",
            True,
            id="Owner",
        ),
        pytest.param(
            "anyscale-admin-editor@anyscale-bridge-deadbeef.iam.gserviceaccount.com",
            True,
            id="Editor",
        ),
        pytest.param(
            "anyscale-admin-editor-not-signer@anyscale-bridge-deadbeef.iam.gserviceaccount.com",
            False,
            id="EditorNotSigner",
        ),
        pytest.param(
            "nope@anyscale-bridge-deadbeef.iam.gserviceaccount.com",
            False,
            id="NotGranted",
        ),
    ],
)
def test_verify_gcp_access_service_account(
    setup_mock_server, service_account: str, expected_result: bool
):
    factory, tracker = setup_mock_server
    tracker.reset(
        {
            ".*": [
                ("200 OK", "access_service_account_permissions.json"),
                ("200 OK", "project_iam_binding_access.json"),
            ]
        }
    )
    assert (
        _verify_gcp_access_service_account(
            factory,
            {**_MOCK_RESOURCES, "anyscale_access_service_account": service_account},
            _gcp_logger(),
        )
        == expected_result
    )


def test_verify_gcp_access_service_account_does_not_exist(setup_mock_server):
    factory, tracker = setup_mock_server
    tracker.reset({".*": [("404 Not Found", None)]})
    assert not _verify_gcp_access_service_account(
        factory, _MOCK_RESOURCES, _gcp_logger(),
    )


@pytest.mark.parametrize(
    ("responses", "expected_result"),
    [
        pytest.param({".*": [("404 Not Found", None)]}, False, id="DoesNotExist"),
        pytest.param(
            {
                ".*": [
                    ("200 OK", "dataplane_service_account.json"),
                    ("200 OK", "project_iam_binding_access.json"),
                ]
            },
            True,
            id="DoesNotExist",
        ),
    ],
)
def test_verify_gcp_dataplane_service_account(
    setup_mock_server, responses, expected_result: bool, capsys
):
    factory, tracker = setup_mock_server
    tracker.reset(responses)
    assert (
        _verify_gcp_dataplane_service_account(factory, _MOCK_RESOURCES, _gcp_logger())
        == expected_result
    )

    _, err = capsys.readouterr()
    if expected_result:
        # Check if there is a warning about missing artifact Registry warnings if we pass verification.
        assert re.search("roles/artifactregistry.reader`", err) and re.search(
            "safe to ignore if you are not using", err
        )


def test_verify_gcp_dataplane_service_account_wrong_project(setup_mock_server, capsys):
    factory, tracker = setup_mock_server
    tracker.reset(
        {
            ".*": [
                ("200 OK", "dataplane_service_account.json"),
                ("200 OK", "project_iam_binding_access.json"),
            ]
        }
    )
    assert _verify_gcp_dataplane_service_account(
        factory, {**_MOCK_RESOURCES, "project_id": "something-else"}, _gcp_logger(),
    )

    _, stderr = capsys.readouterr()
    assert (
        "constraints/iam.disableCrossProjectServiceAccountUsage` is not enforced"
        in stderr
    )


def _gcp_logger() -> GCPLogger:
    return GCPLogger(BlockLogger(), "project_abc", yes=True)



from functools import partial
from typing import Any, Dict

from google.api_core.exceptions import NotFound, PermissionDenied
from google.auth.credentials import Credentials
import google.cloud
from google.cloud import compute_v1, resourcemanager_v3  # noqa
from googleapiclient.discovery import build as api_client_build
from googleapiclient.errors import HttpError

from anyscale.cli_logger import BlockLogger
from anyscale.util import confirm
from anyscale.utils.gcp_utils import (
    binding_from_dictionary,
    check_policy_bindings,
    get_application_default_credentials,
)
from anyscale.utils.network_verification import GCP_SUBNET_CAPACITY


class GCPLogger:
    def __init__(self, logger: BlockLogger, project_id: str, yes: bool = False):
        self.internal = logger
        self.project_id = project_id
        self.yes = yes

    def log_resource_not_found_error(self, resource_name: str, resource_id: str):
        self.internal.error(
            f"Could not find {resource_name} with id {resource_id} in project {self.project_id}. Please validate that you're using the correct GCP project and that the resource values are correct."
        )

    def confirm_missing_permission(self, error_str: str):
        self.internal.error(error_str)
        confirm(
            "If the serivce account has these required permissions granted via other\n"
            "roles, press 'y' to continue with verification or 'N' to abort.",
            yes=self.yes,
        )


class GoogleCloudClientFactory:
    """Factory to generate both Google Cloud Client libraries & Google API Client libraries.

    Google Cloud Client libraries are instantiated by:
    ```
        factory = GoogleCloudClientFactory(credentials=AnonymousCredentials())
        client = factory.compute_v1.ExampleClient()
    ```

    Google API Client libraries are instantiated by:
    ```
        factory = GoogleCloudClientFactory(credentials=AnonymousCredentials())
        client = factory.build("iam", "v1")
    ```
    """

    def __init__(self, credentials: Credentials, force_rest=False, **kwargs):
        kwargs["credentials"] = credentials
        self.kwargs = kwargs
        self.force_rest = force_rest

    def __getattr__(self, client_library: str):
        """Get a wrapped Google Cloud Client library that injects default values from the factory."""
        module = getattr(google.cloud, client_library)
        kwargs = self.kwargs
        if self.force_rest:
            kwargs["transport"] = "rest"

        class WrappedClient:
            def __getattr__(self, client_type: str):
                return partial(getattr(module, client_type), **kwargs)

        return WrappedClient()

    def build(self, service_name: str, version: str):
        """Return a Google API Client with default values from the factor"""
        return api_client_build(
            service_name, version, cache_discovery=False, **self.kwargs
        )


def verify_gcp(resources: Dict[str, Any], logger: BlockLogger) -> bool:
    credentials, credentials_project = get_application_default_credentials(logger)
    specified_project = resources["project_id"]
    if credentials_project != specified_project:
        logger.warning(
            f"Default credentials are for {credentials_project}, but this cloud is being configured for {specified_project}"
        )

    factory = GoogleCloudClientFactory(credentials=credentials)
    gcp_logger = GCPLogger(logger, specified_project)

    return all(
        [
            _verify_gcp_project(factory, resources, gcp_logger),
            _verify_gcp_access_service_account(factory, resources, gcp_logger),
            _verify_gcp_dataplane_service_account(factory, resources, gcp_logger),
            _verify_gcp_networking(factory, resources, gcp_logger),
        ]
    )


def _verify_gcp_networking(
    factory: GoogleCloudClientFactory, resources: Dict, logger: GCPLogger,
) -> bool:
    """Verify the existence and connectedness of the VPC & Subnet."""

    project = resources["project_id"]
    vpc_name = resources["vpc_name"]
    # TODO Verify Internet Gateway
    try:
        vpc = factory.compute_v1.NetworksClient().get(project=project, network=vpc_name)
    except NotFound:
        logger.log_resource_not_found_error("VPC", vpc_name)
        return False

    subnet_name = resources["subnet_name"]
    try:
        subnet = factory.compute_v1.SubnetworksClient().get(
            project=project, subnetwork=subnet_name, region=resources["region"]
        )
    except NotFound:
        logger.log_resource_not_found_error("Subnet", subnet_name)
        return False

    if subnet.network != vpc.self_link:
        logger.internal.error(f"Subnet {subnet_name} is not part of {vpc_name}!")
        return False

    return _gcp_subnet_has_enough_capacity(subnet, logger.internal)


def _gcp_subnet_has_enough_capacity(
    subnet: compute_v1.types.compute.Subnetwork, logger: BlockLogger
) -> bool:
    """Verify if the subnet provided has a large enough IP address block."""
    if GCP_SUBNET_CAPACITY.verify_network_capacity(
        cidr_block_str=subnet.ip_cidr_range, resource_name=subnet.name, logger=logger,
    ):
        logger.info(f"Subnet {subnet.name} verification succeeded.")
        return True
    return False


def _verify_gcp_project(
    factory: GoogleCloudClientFactory, resources, logger: GCPLogger
) -> bool:
    """Verify if the project exists and is active.

    NOTE: This also checks that Compute Engine Service Agent is configured in the default way.
    """
    project_client = factory.resourcemanager_v3.ProjectsClient()
    project_id = resources["project_id"]
    try:
        project = project_client.get_project(name=f"projects/{project_id}")
    except (NotFound, PermissionDenied):
        logger.log_resource_not_found_error("Project", project_id)
        return False
    if project.state != project.State.ACTIVE:
        logger.internal.error(
            f"Project {project_id} is in state: {project.state}, not active"
        )
        return False

    iam_policies = project_client.get_iam_policy(resource=f"projects/{project_id}")
    project_number = project.name.split("/")[1]

    if not check_policy_bindings(
        iam_policies.bindings,
        f"serviceAccount:service-{project_number}@compute-system.iam.gserviceaccount.com",
        {"roles/compute.serviceAgent"},
    ):
        logger.internal.warning(
            "The Compute Engine Service Agent does not have the standard IAM Role of 'roles/compute.serviceAgent' in your project.\n"
            "This is not recommended by Google and may result in an inability for Compute Engine to function properly:\n"
            "https://cloud.google.com/compute/docs/access/service-accounts#compute_engine_service_account"
        )

    # TODO: Verify that APIs are Enabled. This will implicitly be verified during other verification steps.
    return True


def _verify_gcp_access_service_account(
    factory: GoogleCloudClientFactory, resources, logger: GCPLogger
) -> bool:
    """Verify if the Service Account meant to grant Anyscale access to this cloud has access to the specified project.

    NOTE: We verify that this service account can call signBlob on itself because this is necessary for downloading logs.
    """
    project_id = resources["project_id"]
    anyscale_access_service_account = resources["anyscale_access_service_account"]

    service_account_client = factory.build("iam", "v1").projects().serviceAccounts()
    try:
        service_account_iam_policy = service_account_client.getIamPolicy(
            resource=f"projects/-/serviceAccounts/{anyscale_access_service_account}"
        ).execute()
    except HttpError as e:
        if e.status_code == 404:
            logger.log_resource_not_found_error(
                "Anyscale Access Service Account", anyscale_access_service_account
            )
            return False

    if not check_policy_bindings(
        binding_from_dictionary(service_account_iam_policy["bindings"]),
        f"serviceAccount:{anyscale_access_service_account}",
        {"roles/iam.serviceAccountTokenCreator"},
    ):
        logger.confirm_missing_permission(
            f"Service Account {anyscale_access_service_account} must have the `iam.serviceAccounts.signBlob` permission to perform log download from the Anyscale platform.\n"
            "Please grant the `roles/iam.serviceAccountTokenCreator` role to this Service Account to pass this check."
        )
        return False

    project_client = factory.resourcemanager_v3.ProjectsClient()
    iam_policies = project_client.get_iam_policy(resource=f"projects/{project_id}")
    if not check_policy_bindings(
        iam_policies.bindings,
        f"serviceAccount:{anyscale_access_service_account}",
        {"roles/editor", "roles/owner"},
    ):
        logger.confirm_missing_permission(
            f"Service Account {anyscale_access_service_account} must have either `roles/editor` or `roles/owner` roles on the Project {project_id}.\n"
            "A more fine-grained set of permissions may be possible, but may break Anyscale's functionality."
        )
        return False

    return True


def _verify_gcp_dataplane_service_account(
    factory: GoogleCloudClientFactory, resources, logger: GCPLogger
) -> bool:
    """Verify that the Service Account for compute instances exists *in* the specified project.

    Compute Engine's ability to use this role
    This relies on the fact that Compute Engine Service Agent has the roles/compute.serviceAgent Role
    """
    service_account = resources["dataplane_service_account"]
    service_account_client = factory.build("iam", "v1").projects().serviceAccounts()
    try:
        resp = service_account_client.get(
            name=f"projects/-/serviceAccounts/{service_account}"
        ).execute()
    except HttpError as e:
        if e.status_code == 404:
            logger.log_resource_not_found_error(
                "Dataplane Service Account", service_account
            )
            return False
        raise e

    cloud_project = resources["project_id"]
    if resp["projectId"] != cloud_project:
        logger.internal.warning(
            "Service Account {} is in project {}, not {}. Please ensure that `constraints/iam.disableCrossProjectServiceAccountUsage` is not enforced.\n"
            "See: https://cloud.google.com/iam/docs/attach-service-accounts#enabling-cross-project for more information.".format(
                service_account, resp["projectId"], cloud_project
            )
        )

    project_client = factory.resourcemanager_v3.ProjectsClient()
    project_iam_bindings = project_client.get_iam_policy(
        resource=f"projects/{cloud_project}"
    )
    if not check_policy_bindings(
        project_iam_bindings.bindings,
        f"serviceAccount:{service_account}",
        {"roles/artifactregistry.reader", "roles/viewer"},
    ):
        logger.internal.warning(
            f"The dataplane Service Account {service_account} does not have `roles/artifactregistry.reader` or `roles/viewer` on {cloud_project}.\n"
            "This is safe to ignore if you are not using your own container images or are configuring access a different way."
        )

    return True



import subprocess
from typing import Dict, List, Optional, Set, Tuple, Union

import google.auth
from google.auth.credentials import Credentials
from google.iam.v1.policy_pb2 import Binding

from anyscale.cli_logger import BlockLogger


def get_application_default_credentials(
    logger: BlockLogger,
) -> Tuple[Credentials, Optional[str]]:
    """Get application default credentials, or run `gcloud` to try to log in."""
    try:
        return google.auth.default()
    except google.auth.exceptions.DefaultCredentialsError as e:
        logger.warning(
            "Could not automatically determine Google Application Default Credentials, trying to authenticate via GCloud"
        )
        auth_login = subprocess.run(["gcloud", "auth", "application-default", "login"])
        if auth_login.returncode != 0:
            raise RuntimeError("Failed to authenticate via gcloud") from e

        return google.auth.default()


def binding_from_dictionary(
    inp: List[Dict[str, Union[List[str], str]]]
) -> List[Binding]:
    return [Binding(role=b["role"], members=b["members"]) for b in inp]


def check_policy_bindings(
    iam_policy: List[Binding], member: str, possible_roles: Set[str]
) -> bool:
    return any(
        policy.role in possible_roles and member in policy.members
        for policy in iam_policy
    )
